#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 20:52:20 2022

@author: liyuzhe
"""
import scanpy as sc
import numpy as np
from scipy.spatial import distance
import networkx as nx


# Construct adj graph of image-based data 
def graph_computing(adj_coo, cell_num,k):
    edgeList = []
    for node_idx in range(cell_num):
        tmp = adj_coo[node_idx, :].reshape(1, -1)
        distMat = distance.cdist(tmp, adj_coo, 'euclidean')
        res = distMat.argsort()[:k+1]
        tmpdist = distMat[0, res[0][1:k+1]]
        boundary = np.mean(tmpdist) + np.std(tmpdist)
        for j in np.arange(1, k+1):
            if distMat[0, res[0][j]] <= boundary:
                weight = 1.0
            else:
                weight = 0.0
            edgeList.append((node_idx, res[0][j], weight))

    return edgeList


def edgeList2edgeDict(edgeList, nodesize):
    graphdict = {}
    tdict = {}
    for edge in edgeList:
        end1 = edge[0]
        end2 = edge[1]
        tdict[end1] = ""
        tdict[end2] = ""
        if end1 in graphdict:
            tmplist = graphdict[end1]
        else:
            tmplist = []
        tmplist.append(end2)
        graphdict[end1] = tmplist

    # check and get full matrix
    for i in range(nodesize):
        if i not in tdict:
            graphdict[i] = []

    return graphdict


def graph_construction(adj_coo, cell_N, k):
    adata_Adj = graph_computing(adj_coo, cell_N, k)
    graphdict = edgeList2edgeDict(adata_Adj, cell_N)
    adj_org = nx.adjacency_matrix(nx.from_dict_of_lists(graphdict))
    
    return adj_org




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 19:58:54 2022

@author: liyuzhe
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GATv2Conv



class GAT_Encoder(nn.Module):
    def __init__(self, in_channels,num_heads, hidden_dims, dropout, concat):
        super(GAT_Encoder, self).__init__()
        
        self.num_heads = num_heads
        self.in_channels = in_channels
        self.latent_dim = 10
        self.conv = GATv2Conv

        self.hidden_layer1 = self.conv(
            in_channels=in_channels, out_channels=hidden_dims[0],
            heads=self.num_heads['first'],
            dropout=dropout[0],
            concat=concat['first'])
        in_dim_hidden = hidden_dims[0] * self.num_heads['first'] if concat['first'] else hidden_dims[0]

        self.hidden_layer2 = self.conv(
            in_channels=in_dim_hidden, out_channels=hidden_dims[1],
            heads=self.num_heads['second'],
            dropout=dropout[1],
            concat=concat['second'])
        in_dim_final = hidden_dims[-1] * self.num_heads['second'] if concat['second'] else hidden_dims[-1]
        

        self.conv_z = self.conv(in_channels=in_dim_final, out_channels=self.latent_dim,
                                heads=self.num_heads['mean'], concat=False, dropout=0.2)

    def forward(self, x, edge_index):
        hidden_out1, attn_w_1 = self.hidden_layer1(x, edge_index, return_attention_weights=True)
        hidden_out1 = F.leaky_relu(hidden_out1)
        hidden_out2, attn_w_2 = self.hidden_layer2(hidden_out1, edge_index, return_attention_weights=True)
        hidden_out2 = F.leaky_relu(hidden_out2)
        hidden_out2 = F.dropout(hidden_out2, p=0.4, training=self.training)
        last_out = hidden_out2
        z, attn_w_z = self.conv_z(last_out, edge_index, return_attention_weights=True)
       
        return z , (attn_w_1, attn_w_2, attn_w_z)
    

class InnerProduct(nn.Module):
    """Decoder for using inner product for prediction."""
    def __init__(self,act=torch.relu):
        super(InnerProduct, self).__init__()
        self.act = act

    def forward(self, z):
        recon_a = self.act(torch.mm(z, z.t()))
        return recon_a 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Mar 13 15:09:45 2022

@author: liyuzhe
"""
import os
import numpy as np
import torch
from tqdm.auto import tqdm


def adjust_learning_rate(init_lr, optimizer, iteration,seperation):
    lr = max(init_lr * (0.9 ** (iteration//seperation)), 0.0001)
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr
    return lr


class EarlyStopping:
    """
    Early stops the training if loss doesn't improve after a given patience.
    """
    def __init__(self, patience=10, verbose=False, checkpoint_file=''):
        """
        Parameters
        ----------
        patience 
            How long to wait after last time loss improved. Default: 10
        verbose
            If True, prints a message for each loss improvement. Default: False
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.loss_min = np.Inf
        self.checkpoint_file = checkpoint_file

    def __call__(self, loss, model):
        # loss=loss.cpu().detach().numpy()
        if np.isnan(loss):
            self.early_stop = True
        score = -loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(loss, model)
        elif score < self.best_score:
            self.counter += 1
            if self.verbose:
                print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter > self.patience:
                self.early_stop = True
                model.load_model(self.checkpoint_file)
        else:
            self.best_score = score
            self.save_checkpoint(loss, model)
            self.counter = 0

    def save_checkpoint(self, loss, model):
        '''
        Saves model when loss decrease.
        '''
        if self.verbose:
            print(f'Loss decreased ({self.loss_min:.6f} --> {loss:.6f}).  Saving model ...')
        torch.save(model.state_dict(), self.checkpoint_file)
        self.loss_min = loss



def train_SPACE(model, train_data, outdir, epoch=5000,lr=0.005, a=0.5,loss_type='MSE',patience=50, GPU=0,seed=42, verbose=False):
    
    np.random.seed(seed)
    if torch.cuda.is_available(): # cuda device
        device='cuda'
        torch.cuda.set_device(GPU)
        torch.cuda.manual_seed_all(seed)
    else:
        device='cpu'
    
    torch.manual_seed(seed)
    model.to(device)
    model.train()
    early_stopping = EarlyStopping(patience=patience, checkpoint_file=os.path.join(outdir,'model.pt'))

    for epoch in tqdm(range(0, epoch+1)):
        epoch_loss = 0.0
        optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=5e-4)
        epoch_lr=adjust_learning_rate(lr, optimizer, epoch, seperation=50)
            
        x, edge_index = train_data.x.to(torch.float).to(device), train_data.edge_index.to(torch.long).to(device) 
        optimizer.zero_grad()
            
        z, _ = model.encode(x, edge_index) 
        graph_loss = model.graph_loss(z, train_data.pos_edge_label_index) * a  
        loss = graph_loss  
            
        reconstructed_features = model.decoder_x(z)
        if loss_type=='BCE': 
            feature_loss = torch.nn.functional.binary_cross_entropy(reconstructed_features, x) * 10
        elif loss_type=='MSE':
            feature_loss = torch.nn.functional.mse_loss(reconstructed_features, x) * 10
            
        loss += feature_loss
        loss.backward()
        optimizer.step()
        epoch_loss += loss.item()
            
        # if epoch%50==0:
        #     if verbose:
        #         print('Epoch {:03d} -- Total epoch loss: {:.4f} -- Feature decoder epoch loss: {:.4f} -- Graph decoder epoch loss: {:.4f} -- epoch lr: {:.4f}'.format(epoch, epoch_loss, feature_loss, epoch_loss-feature_loss, epoch_lr))
        #     else:
        #         print('====> Epoch: {}, Loss: {:.4f}'.format(epoch, epoch_loss)) 
        # early_stopping(epoch_loss, model)
        # if early_stopping.early_stop:
        #     print('EarlyStopping: run {} iteration'.format(epoch))
        #     break
        if verbose:
            if epoch%50==0:
                print('====> Epoch: {}, Loss: {:.4f}'.format(epoch, epoch_loss)) 
        early_stopping(epoch_loss, model)
        if early_stopping.early_stop:
            print('====> Epoch: {}, Loss: {:.4f}'.format(epoch, epoch_loss)) 
            print('EarlyStopping: run {} iteration'.format(epoch))
            break
    
    return device



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Mar 13 16:21:44 2022

@author: liyuzhe
"""
import os
import random
import numpy as np
import scanpy as sc
import scipy as sci
import networkx as nx
from sklearn.preprocessing import MaxAbsScaler

import torch
import torch_geometric
import torch_geometric.transforms as T
from torch_geometric.data import Data

from .layer import GAT_Encoder
from .model import SPACE_Graph
from .train import train_SPACE
from .utils import graph_construction



def SPACE(adata,k=20,alpha=0.5,seed=42,GPU=0,epoch=5000,lr=0.005,patience=50,outdir='./',loss_type='MSE',save_attn=False,verbose=False,heads=6):
    """
    Single-Cell integrative Analysis via Latent feature Extraction
    
    Parameters
    ----------
    adata
        An AnnData Object with spot/cell x gene matrix stored.
    k
        The number cells consider when constructing Adjacency Matrix. Default: 20.
    alpha
        The relative ratio of reconstruction loss of Adjacency Matrix. Default: 0.5 (0.5 is recommanded for domain detection, 0.05 is recommanded for cell type detection).
    lr
        Learning rate. Default: 5e-3.
    patience
        Patience in early stopping. Default: 50.
    epoch
        Max epochs for training. Default: 5000.
    loss_type
        Loss Function of feature matrix reconstruction loss. Default: MSE (BCE was recommanded for Visium data's domain finding). 
    GPU
        Index of GPU to use if GPU is available. Default: 0.
    outdir
        Output directory. Default: '.'.
    save_attn
        Save attention weights, True or False. Default: False.
    verbose
        Verbosity, True or False. Default: False.
    heads
        Number of attention heads. Default: 6.
    
    
    Returns
    -------
    adata with the low-dimensional representation of the data stored at adata.obsm['latent'], and calculated neighbors and Umap. 
    The output folder contains:
    adata.h5ad
        The AnnData Object with the low-dimensional representation of the data stored at adata.obsm['latent'].
    checkpoint
        model.pt contains the variables of the model.
    """
    from torch_geometric import seed_everything
    np.random.seed(seed) 
    torch.manual_seed(seed)
    random.seed(seed)
    seed_everything(seed)
    os.makedirs(outdir, exist_ok=True)
        
    print('Construct Graph')
    graph_dict = graph_construction(adata.obsm['spatial'], adata.shape[0],k=k)
    adj = graph_dict.toarray()
    print('Average links: {:.2f}'.format(np.sum(adj>0)/adj.shape[0]))

    G = nx.from_numpy_array(adj).to_undirected() 
    edge_index = (torch_geometric.utils.convert.from_networkx(G)).edge_index
    
    if sci.sparse.issparse(adata.X):
        X_hvg = adata.X.toarray()
    else:
        X_hvg = adata.X.copy()
        
    if loss_type == 'BCE':
        scaler = MaxAbsScaler() # MinMaxScaler() 
        scaled_x = torch.from_numpy(scaler.fit_transform(X_hvg))
    elif loss_type == 'MSE':
        scaled_x = torch.from_numpy(X_hvg) 
    
    #prepare training data
    data_obj = Data(edge_index=edge_index, x=scaled_x) 
    data_obj.num_nodes = X_hvg.shape[0] 
    data_obj.train_mask = data_obj.val_mask = data_obj.test_mask = data_obj.y = None

    transform = T.RandomLinkSplit(num_val=0.0, num_test=0.0, is_undirected=True, 
                                  add_negative_train_samples=False, split_labels=True)
    train_data, _ , _ = transform(data_obj) 
    num_features = data_obj.num_features
    
    print('Load SPACE model')
    encoder = GAT_Encoder(
        in_channels=num_features,
        num_heads={'first':heads,'second':heads,'mean':heads},
        hidden_dims=[128,128],
        dropout=[0.3,0.3],
        concat={'first': True, 'second': True})
    model = SPACE_Graph(encoder= encoder,decoder=None,loss_type=loss_type)
    print(model) 
    print('Train SPACE model')
    
    device=train_SPACE(model, train_data, epoch=epoch, lr=lr, patience=patience, GPU=GPU, seed=seed, a=alpha,loss_type=loss_type, outdir= outdir, verbose=verbose)
    
    # Load model
    pretrained_dict = torch.load(os.path.join(outdir,'model.pt'), map_location=device)                            
    model_dict = model.state_dict()
    pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in model_dict}
    model_dict.update(pretrained_dict) 
    model.load_state_dict(model_dict)
    model = model.eval()

    node_embeddings = []
    x, edge_index = data_obj.x.to(torch.float).to(device), data_obj.edge_index.to(torch.long).to(device)
    z_nodes, attn_w = model.encode(x, edge_index)
    node_embeddings.append(z_nodes.cpu().detach().numpy()) 
    node_embeddings = np.array(node_embeddings) 
    node_embeddings = node_embeddings.squeeze()

    adata.obsm['latent']=node_embeddings 
    sc.pp.neighbors(adata, n_neighbors=20, n_pcs=10,use_rep='latent',random_state=seed,key_added='SPACE') 
    sc.tl.umap(adata,random_state=seed, neighbors_key='SPACE')
    
    # Save attention weights
    if save_attn:
        for i in range(2):
            edge_index, attention_weights = attn_w[i]
            edge_index, attention_weights = edge_index.detach().cpu(), attention_weights.detach().cpu()
            
            layer_filename = f'GAT_Layer_{i+1}_edge_index.pt'
            edges_filepath = os.path.join(outdir, layer_filename)
            layer_filename = f'GAT_Layer_{i+1}_attention_weights.pt'
            attn_w_filepath = os.path.join(outdir, layer_filename)
            
            torch.save(edge_index, edges_filepath)
            torch.save(attention_weights, attn_w_filepath)

    #Output
    adata.write(os.path.join(outdir,'adata.h5ad'))   
    return adata
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 20:33:19 2022

@author: liyuzhe
"""
import torch
from torch.nn import Sequential, Linear, ReLU, BatchNorm1d, Dropout, Sigmoid, LeakyReLU
from torch_geometric.nn import GAE, InnerProductDecoder
from torch_geometric.utils import (negative_sampling, remove_self_loops, add_self_loops)

EPS = 1e-15


class SPACE_Graph(GAE):
    def __init__(self, encoder, decoder,loss_type='MSE'):
        super(SPACE_Graph, self).__init__(encoder, decoder)
        
        self.decoder = InnerProductDecoder()
        
        if loss_type == 'BCE':
            self.decoder_x = Sequential(Linear(in_features=self.encoder.latent_dim, out_features=128),
                                      BatchNorm1d(128),
                                      LeakyReLU(),
                                      Dropout(0.1),
                                      Linear(in_features=128, out_features=self.encoder.in_channels),
                                      Sigmoid())
        else:
            self.decoder_x = Sequential(Linear(in_features=self.encoder.latent_dim, out_features=128),
                                      BatchNorm1d(128),
                                      LeakyReLU(),
                                      Dropout(0.1),
                                      Linear(in_features=128, out_features=self.encoder.in_channels),
                                      ReLU())

    
    def encode(self, *args, **kwargs):
        
        z, attn_w = self.encoder(*args, **kwargs)
        
        return z, attn_w


    def graph_loss(self, z, pos_edge_index, neg_edge_index=None):

        self.decoded = self.decoder(z, pos_edge_index, sigmoid=True)
        pos_loss = -torch.log(self.decoded + EPS).mean()

        # Do not include self-loops in negative samples
        pos_edge_index, _ = remove_self_loops(pos_edge_index)
        pos_edge_index, _ = add_self_loops(pos_edge_index)
        if neg_edge_index is None:
            neg_edge_index = negative_sampling(pos_edge_index, z.size(0))
        neg_loss = -torch.log(1 - self.decoder(z, neg_edge_index, sigmoid=True) + EPS).mean()

        return pos_loss + neg_loss
    
     
    def load_model(self, path):

        pretrained_dict = torch.load(path, map_location=lambda storage, loc: storage)                            
        model_dict = self.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in model_dict}
        model_dict.update(pretrained_dict) 
        self.load_state_dict(model_dict)
    
